install_requires = [
    "attrs",
    "matplotlib",
    "numpy >= 1.12",
    "pandas",
    "pymedphys_databases >= 0.10.0, < 0.11.0",
    "pymedphys_fileformats >= 0.10.0, < 0.11.0",
    "pymedphys_mudensity >= 0.10.0, < 0.11.0",
    "pymedphys_utilities >= 0.10.0, < 0.11.0",
    "python_dateutil"
]
